/* Copyright (C) 1997 Tom Wheeley, see file COPYING for details */

/*
 * ARGS.C  Argument handling functions for -o, -i and -l switches
 *         -x=format,arg1,arg2,arg3
 *         -x,arg1,arg2
 */

#define _CAS_ARGS_C_

#include "cas.h"
#include "misc.h"
#include "token.h"
#include "args.h"

/* casrc loading functions */ 

void args_load_defaults(void)
{
  FILE *rc=NULL;
  char *filename=NULL;
  char *tmp=NULL;
  char *linedata=NULL;
  char *linectrl=NULL;
  char buffer[ARGS_RC_LINELEN];
  int macro;

  defaults = tsys_calloc(sizeof (defaults_t));

  /* initialise macro linked-list */
  defaults->macro = tsys_malloc(sizeof (macro_t));
  defaults->macro->name = NULL;
  defaults->macro->args = NULL;
  defaults->macro->next = NULL;

  /* open correct file.  Not as easy as you might think... */

  #ifdef CAS_DOS

  /* for DOS, we look in the same directory as cas.exe.  argv0 contains the
   * fully qualified file name of cas.exe.  We need to use CAS_DIRSEP as
   * for djgpp the directory separator is `/' not `\'.
   */
  
  filename = tsys_malloc(strlen(argv0) + strlen(CAS_DOS_RC));
  strcpy(filename, argv0);
  if ((tmp = strrchr(filename, CAS_DIRSEP)) == NULL)
    tmp = filename;
  else
    ++tmp;
  strcpy(tmp, CAS_DOS_RC);
  rc = tsys_fopen(filename, "r", FALSE);
  free(filename);

  #else		/* CAS_DOS */
  # ifdef CAS_UNIX

  /* for Unix, we first look in the user's home directory, if nothing
     there then look for the system-wide rc file */
     
  filename = tsys_malloc(FILENAME_MAX*2+1);		/* big enough... */
  if (getenv("HOME") != NULL) {
    strcpy(filename, getenv("HOME"));
    strcat(filename, "/");
    strcat(filename, CAS_UNIX_RC);
    if ((rc = tsys_fopen(filename,"r",FALSE)) == NULL) {
      Msg(M_DEBUG,"casrc","%s not found, using %s",filename,
    				CAS_UNIX_RCDIR CAS_UNIX_RC);
    }
  } else {
    Msg(M_DEBUG,"casrc","${HOME} not set, cannot locate ~/.casrc");
  }
  if (rc == NULL)
    rc = tsys_fopen(CAS_UNIX_RCDIR CAS_UNIX_RC, "r", FALSE);
  free(filename);

  # else	/* CAS_UNIX */

  /* For other systems, just look 'here'.  Wherever 'here' is. */

  rc = tsys_fopen(CAS_EXOTIC_RC, "r", FALSE);

  # endif	/* CAS_UNIX */
  #endif	/* CAS_DOS */

  if (!rc) {
    Msg(M_DEBUG,"casrc","Unable to load rc file");
    return;
  }

  /* read defaults */

  for(;;) {
    fgets(buffer,ARGS_RC_LINELEN,rc);
    if (feof(rc)) break;
	/* remove trailing newline */
    if ((tmp=strchr(buffer,'\n')) !=NULL)
      *tmp='\0';
    /* remove spaces at beginning */
    for(tmp=buffer;isspace(*tmp);tmp++);
    if (*tmp=='#' || *tmp==';' || *tmp=='\0') continue;

    /* if a macro, note the fact and skip the text `macro' */
    macro=0;
    if (!strnicmp(tmp,"macro",5)) {
      macro=1;
      for(tmp+=5;isspace(*tmp);tmp++);
    }
      
    /* split first word (separated by *any* form of whitespace) away from
       the rest of the line.  */
    linectrl=tmp;
    for(linedata=linectrl;!isspace(*linedata) && *linedata!='\0';++linedata)
      ;
    *linedata='\0';
    for(++linedata;isspace(*linedata);++linedata)
      ;
    
	/* call appropriate function to add information to wherever */
	if (macro) {
	  add_macro(linectrl, linedata);
	} else {
	  tmp=strchr(linectrl,'.');
	  if (tmp) {
        *tmp++='\0';
        add_arg(linectrl, tmp, linedata);
      } else {
        add_setting(linectrl, linedata);
      }
    }
  }

  tsys_fclose(rc);
}

/* Perhaps the arg checking in main() should be subcontracted so that it can
 * be called from here.  Then all long options could be settings in casrc. 
 * This would allow settings such as `verbose' and `debug=file', as well as
 * maybe the perhaps less useful `output' etc.
 */
static void add_setting(char *name, char *data)
{
  if (!stricmp(name,"model")) {		/* code duplication with main.c here */
    if (!strnicmp(data,"cfx",3)) data+=3;
    if (!strnicmp(data,"fx",2)) data+=2;
    if (!strcmp(data,"7700") || !strcmp(data,"7")) {
      flags.model=MDL_7700; return; }
    if (!strcmp(data,"9700") || !strcmp(data,"9")) {
      flags.model=MDL_9700; return; }
    if (!strcmp(data,"9750") ) {
      flags.model=MDL_9750; return; }
    if (!strcmp(data,"9800") || !strcmp(data,"8")) {
      flags.model=MDL_9800; return; }
    if (!strcmp(data,"9850") || !strcmp(data,"5")) {
      flags.model=MDL_9850; return; }
    if (!strcmp(data,"9950") ) {
      flags.model=MDL_9950; return; }
    if (!stricmp(data,"any") || !strcmp(data,"*")) {
      flags.model=MDL_UNKNOWN; return; }
    Msg(M_DEBUG,"casrc","unknown model: `%s'",data);
  }

  Msg(M_DEBUG,"casrc","unknown identifier '%s'",name);
}


static void add_arg(char *iol, char *type, char *data)
{
  int whicharr = ARGS_DEF_UNDEFINED;
  int index;
  
  if (!stricmp(iol,"in")) whicharr=ARGS_DEF_IN;
  if (!stricmp(iol,"out")) whicharr=ARGS_DEF_OUT;
  if (!stricmp(iol,"list")) whicharr=ARGS_DEF_LIST;
  if (whicharr==ARGS_DEF_UNDEFINED) {
    Msg(M_DEBUG,"casrc","default arg not tied to in, out or list: `%s.%s'",
   			iol,type);
   	return;
  }

  if (whicharr==ARGS_DEF_LIST) {
    index=string_to_type(type);
  } else {
    index=string_to_format(type);
  }
  
  if (index==0) {
    Msg(M_DEBUG,"casrc","default arg does not match a type/format: `%s.%s'",
    		iol,type);
    return;
  }
  
  switch(whicharr) {
    case ARGS_DEF_IN:
      defaults->in[index] = args_split(data, defaults->in[index]);
      break;
    case ARGS_DEF_OUT:
      defaults->out[index] = args_split(data, defaults->out[index]);
      break;
    case ARGS_DEF_LIST:
      defaults->list[index] = args_split(data, defaults->list[index]);
      break;
  }
}


static void add_macro(char *name, char *data)
{
  macro_t *newmac = NULL;
  macro_t *mac = NULL;
  macro_t *prev = NULL;
  
  /* to avoid recursive macros, we parse the macros args before adding it to
   * the macro list.  Thus the macro's name is untouched if present in the
   * macro's argument list */
  newmac = tsys_malloc(sizeof (macro_t));
  newmac->name = tsys_strdup(name);
  newmac->args = args_split(data, NULL);
  newmac->next=NULL;
  
  prev = defaults->macro;
  for (mac = defaults->macro; mac != NULL; prev = mac, mac = mac->next) {
    if (mac->name == NULL)
      continue;			/* head of list */
    if (!stricmp(name, mac->name)) {
      args_free(mac->args);
      mac->args = newmac->args;
      free(newmac->name);
      free(newmac);
      return;
    }
  }
  prev->next = newmac;  
}


static int args_expand_macro(char *name, arg_t **args)
{
  macro_t *mac;
  arg_t *src;
  
  for (mac = defaults->macro; mac != NULL; mac = mac->next) {
    if (mac->name==NULL)
      continue;
    if (!stricmp(name, mac->name)) {
      /* expand macro */
      for (src = mac->args; src !=NULL; src = src->next) {
        (*args)->next = tsys_malloc(sizeof (arg_t));
        (*args) = (*args)->next;
        (*args)->name = tsys_strdup(src->name);
        (*args)->value = tsys_strdup(src->value);
        (*args)->set = src->set;
        (*args)->next = NULL;
      }
      return 1;
    }
  }
  
  return 0;
}

/* more general args functions */

arg_t *args_split(char *argstr, arg_t *endp)
{
  char *name=NULL;
  char *value=NULL;
  arg_t *argstart=NULL;
  arg_t *args=NULL;

  if (!argstr) return NULL;
  if (*argstr=='='||*argstr==':') argstr++;

  argstart=args=tsys_malloc(sizeof (arg_t));    /* dummy record */
  args->name=NULL;
  args->set=0;
  args->next=endp;

  if (*argstr==',') {
    args->next=tsys_malloc(sizeof (arg_t));
    args=args->next;
    args->name="";
    args->value="";
    args->set=1;
    args->next=endp;
  }

  for(name=strtok(argstr,",");name;name=strtok(NULL,",")) {

    if (args_expand_macro(name, &args))
      continue;
    /* create new arg node */
    args->next=tsys_malloc(sizeof (arg_t));
    
    args=args->next;
    /* initialise new node */
    args->name=NULL;
    args->value="";
    args->set=1;
    args->next=endp;

    /* if a `no-XXX' switch, set set indicator to unset */
    if (strnicmp(name,"no-",3) ==0) {
      name+=3;
      args->set=0;
    }

    value=strchr(name,'=');
    if (value != NULL) {          /* key=value style argument */
      *(value++)='\0';
      if (args->set==0) {
        Msg(M_WARN,"args","keyed value should not start with `no-' in `%s'",
                  name);
      }
    }

    args->name=strdup_trim(name," \t");
    if (value) args->value=strdup_trim(value," \t");
  }
 

  /* remove header from linked list */
  args=argstart->next;
  free(argstart);

  return args;
}

void args_free(arg_t *args)
{


}

char *args_get_first(arg_t **args)
{
  char *fmt;
  arg_t *temp;

  if (!args || !*args) return NULL;
  fmt=(*args)->name;
  if (*((*args)->value)) {
    Msg(M_WARN,"args","format specifier should not be keyed value");
    free((*args)->value);
  }
  if ((*args)->set==0) {
    Msg(M_WARN,"args","format specifier should not be an unset argument");
  }

  temp=(*args);
  (*args)=temp->next;
  free(temp);

  return fmt;
}

void args_join(arg_t *new, int whicharr, int index, darg_t *dargs)
{
  dargs->newargs=new;

  switch (whicharr) {
    case ARGS_DEF_IN:
      dargs->defargs=defaults->in[index];
      break;
    case ARGS_DEF_OUT:
      dargs->defargs=defaults->out[index];
      break;
    case ARGS_DEF_LIST:
      dargs->defargs=defaults->list[index];
      break;
    default:
      Msg(EXIT_BUG,"args","Invalid array specifier in args_join() (%d)",whicharr);
  }
}


static arg_t *get_arg_match(char *name, arg_t *args)
{
  for (;args;args=args->next)
    if (!stricmp(name,args->name))
      return args;

  return NULL;
}

static arg_t *get_darg_match(darg_t dargs, va_list va)
{
  char *cur=NULL;
  arg_t *new=NULL;
  arg_t *def=NULL;

  for(cur=va_arg(va,char *); cur!=NULL; cur=va_arg(va,char *)) {
    new=get_arg_match(cur,dargs.newargs);
    if (new != NULL) break;     /* that will do */
    if (def==NULL) def=get_arg_match(cur,dargs.defargs);
  }

  return (new != NULL) ? new : def;
}


int args_is_set(darg_t dargs, ...)
{
  va_list va;
  arg_t *s;

  va_start(va, dargs);
  s=get_darg_match(dargs,va);
  va_end(va);

  if (s==NULL)
    return 0;
  else
    return (s->set==1) ? 1 : 0;
}


char *args_get_string(darg_t dargs, ...)
{
  va_list va;
  arg_t *s;

  va_start(va, dargs);
  s=get_darg_match(dargs,va);
  va_end(va);

  if (s==NULL)
    return "";
  else
    return s->value;
}


long args_get_long(darg_t dargs, long def, ...)
{
  va_list va;
  arg_t *s=NULL;
  char *e=NULL;
  long l;

  va_start(va, def);
  s=get_darg_match(dargs,va);
  va_end(va);

  l=strtol(s->value,&e,0);
  if (e && *e) {
    Msg(M_WARN,"args","incomplete number `%s' from `%s'",s,e);
    return def;
  }

  return l;
}

char args_get_char(darg_t dargs, char def, ...)
{
  va_list va;
  arg_t *s=NULL;
  char *e=NULL;
  long l;

  va_start(va, def);
  s=get_darg_match(dargs,va);
  va_end(va);

  if (!s) return def;

  if (*s->value=='$') {            /* actual character */
    if (s->value[1] != '\0') {
      return s->value[1];
    } else {
      return def;
    }
  } else {                  /* ascii code of character */
    l=strtol(s->value,&e,0);
    if (e && *e) {
      Msg(M_WARN,"args","incomplete number `%s' from `%s'",s,e);
      return def;
    } else {
      return (char) (l & 255);
    }
  }

  return def;
}

int args_is_set_keyed(darg_t dargs, char *key, ...)
{
  char *tmp=NULL;
  char *mat=NULL;
  arg_t *s;
  va_list va;

  va_start(va, key);
  s=get_darg_match(dargs,va);
  va_end(va);

  if (s != NULL) return s->set;

  mat=args_get_string(dargs,key,NULL);
  va_start(va, key);
  for(;;) {
    tmp=va_arg(va,char *);
    if (tmp==NULL || !stricmp(mat,tmp)) break;
  }
  va_end(va);

  return (tmp==NULL) ? 0 : 1;
}

/* end of args.c */
