/* Copyright (C) 1997 Tom Wheeley, see file COPYING for details */

/*
 * FBMP.C  Routines for reading/writing Windows 3.0 RGB Bitmap files
 */
 
#define _CAS_FBMP_C_

#include "cas.h"
#include "misc.h"
#include "args.h"
#include "token.h"
#include "list.h"
#include "fbmp.h"

void BMPReadData(char *filename, casinfo *info)
{
  darg_t dargs;
  FILE *bmp=NULL;
  unsigned char *data;
  int x,y,z;
  unsigned char chunk;
  int flg_inv;
  int padsize;
  casinfo *curblk;

  args_join(flags.input.args,ARGS_DEF_IN,info->type,&dargs);

  if (!(bmp=tsys_fopen(filename,"rb",FALSE))) {
    Msg(M_ERROR,"bmp/read","Unable to open `%s'",filename);
    return;
  }

  if ((curblk=BMPReadHeader(bmp)) == NULL) {
    Msg(M_ERROR,"bmp/read","Error reading header");
    tsys_fclose(bmp);
    return;
  }

  if (curblk->type == tSSCol) {
#ifdef PROPER_SPELLING
    Msg(M_ERROR,"bmp.read","Colour is currently not supported");
#else
    Msg(M_ERROR,"bmp.read","Color is currently not supported");
#endif
    free(curblk);
    return;
  }

  flg_inv=args_is_set(dargs,"inverse","inv",NULL);
  data = curblk->data = tsys_malloc(curblk->s.ss.width * curblk->s.ss.height);
  padsize=((curblk->s.ss.width - 1) & ~3) + 4 - curblk->s.ss.width;
  Msg(M_DEBUG,"bmp.read","padding: %d",padsize);

  for (y=curblk->s.ss.height-1;y>=0;y--) {
    for (x=0;x<curblk->s.ss.width;x+=8) {
      fread(&chunk,1,1,bmp);
      for (z=0;z<8;z++)
        data[SSOFFSET(curblk,x+z,y)]=((chunk & pow2[7-z])!=0)?!flg_inv:flg_inv;
    }
    for (x=0;x<padsize;x++)       /* read padding */
      fread(&chunk,1,1,bmp);
  }
  tsys_fclose(bmp);

  curblk->length=curblk->s.ss.width * curblk->s.ss.height;
  info->next=curblk;
}


casinfo *BMPReadHeader(FILE *bmp)
{
  A_BITMAPFILEHEADER bmfh;
  A_BITMAPINFOHEADER bmih;
  UINT btLE='B'+('M'<<8);
  UINT btBE=('B'<<8)+'M';
  casinfo *info;

  fread(&bmfh,1,sizeof(bmfh),bmp);
  fread(&bmih.biSize,1,sizeof(BITMAPINFOHEADER),bmp);
  fseek(bmp,bmfh.bfOffBits,SEEK_SET);

  Msg(M_DEBUG,"bmp.readhdr","width: %ld, height: %ld, bpp: %d",
                bmih.biWidth,bmih.biHeight,bmih.biBitCount);

  if (bmfh.bfType!=btLE && bmfh.bfType!=btBE) {
    Msg(M_ERROR,"bmp.readhdr","Not a Bitmap file");
    return NULL;
  }
#if 0
  if (bmih.biHeight>SS_YMAX) {
    Msg(M_ERROR,"bmp.readhdr","Bitmap too high: %ld > %ld",bmih.biHeight,SS_YMAX);
    return NULL;
  }
  if (bmih.biWidth>SS_XMAX) {
    Msg(M_ERROR,"bmp.readhdr","Bitmap too wide: %ld > %ld",bmih.biWidth,SS_XMAX);
    return NULL;
  }
  if (bmih.biBitCount>SS_BMAX) {
    Msg(M_ERROR,"bmp.readhdr","Too many bits-per-pixel: %d > %d",bmih.biBitCount,SS_BMAX);
    return NULL;
  }
#endif

  info=tsys_malloc(sizeof(casinfo));

  info->length=0;
  info->data=NULL;
  info->next=NULL;
  info->s.ss.width=bmih.biWidth;                /* Width */
  info->s.ss.height=bmih.biHeight;              /* Height */
  info->s.ss.bpp=bmih.biBitCount;               /* bpp */
  info->s.ss.order='\0';
  if (info->s.ss.bpp==1)
    info->type=tSSMono;
  else
    info->type=tSSCol;

  return info;
}


void BMPWriteData(char *filename,casinfo *info)
{
  for (;info;info=info->next) {
    switch(info->type) {
      case tSSMono:
        BMPWriteMono(filename,info);
        return;
      case tSSCol:
#ifdef PROPER_SPELLING
        Msg(M_ERROR,"bmp.write","Cannot currently deal with colour screenshots");
#else
        Msg(M_ERROR,"bmp.write","Cannot currently deal with color screenshots");
#endif
      case tText:
      case tDesc:
        break;
      default:
        Msg(M_WARN,"bmp.write","unable to write data of type `%s'",s_types[info->type]);
    }
  }
}


void BMPWriteMono(char *filename, casinfo *info)
{
  darg_t dargs;
  FILE *bmp=NULL;
  int x,y,z;
  unsigned char *data;
  unsigned char chunk;
  int flg_inv;
  int padsize;

  args_join(flags.output.args,ARGS_DEF_OUT,fBMP,&dargs);

  flg_inv=args_is_set(dargs,"inverse","inv",NULL);

  if (!(bmp=tsys_fopen(filename,"wb",FALSE))) return;

  BMPWriteHeader(bmp,info);

  data=info->data;
  padsize=((info->s.ss.width - 1) & ~3) + 4 - info->s.ss.width;

  for (y=info->s.ss.height-1;y>=0;y--) {
    for (x=0;x<info->s.ss.width;x+=8) {
      chunk=0;
      for (z=0;z<8;z++)
        chunk+=((data[SSOFFSET(info,x+z,y)]?!flg_inv:flg_inv)?1:0) * pow2[7-z];
      fwrite(&chunk,1,1,bmp);
    }
    chunk=0;
    for (x=0;x<padsize;x++)       /* write 32 bit padding */
      fwrite(&chunk,1,1,bmp);
  }
  tsys_fclose(bmp);
}


void BMPWriteHeader(FILE *bmp, casinfo *info)
{
  A_BITMAPFILEHEADER bmfh;
  A_BITMAPINFOHEADER bmih;
  RGBQUAD col[1];

  bmfh.bfType='B'+('M' << 8);
  bmfh.bfReserved1=0;bmfh.bfReserved2=0;
  bmfh.bfOffBits = sizeof(bmfh)
                 + sizeof(BITMAPINFOHEADER)
                 + sizeof(RGBQUAD) * 2;
  bmfh.bfSize=bmfh.bfOffBits + info->s.ss.width * info->s.ss.height / 8;

  bmih.biSize=sizeof(BITMAPINFOHEADER);
  bmih.biWidth=info->s.ss.width;
  bmih.biHeight=info->s.ss.height;
  bmih.biBitCount=info->s.ss.bpp;
  bmih.biPlanes=1;
  bmih.biCompression=0;
  bmih.biSizeImage=0;
  bmih.biXPelsPerMeter=0x00000b13;  /* taken from BMP made by CasioWin */
  bmih.biYPelsPerMeter=0x00000b13;
  bmih.biClrUsed=0;
  bmih.biClrImportant=0;

  col[0].rgbBlue=0xff;col[0].rgbGreen=0xff;
  col[0].rgbRed=0xff;col[0].rgbReserved=0;

  col[1].rgbBlue=0;col[1].rgbGreen=0;
  col[1].rgbRed=0;col[1].rgbReserved=0;

  fwrite(&bmfh,1,sizeof(bmfh),bmp);
  fwrite(&bmih.biSize,1,sizeof(BITMAPINFOHEADER),bmp);
  fwrite(&col[0],1,sizeof(RGBQUAD),bmp);
  fwrite(&col[1],1,sizeof(RGBQUAD),bmp);

  Msg(M_DEBUG,"bmp.writehdr","Data offset: %04x",ftell(bmp));
}


/* end of fbmp.c */
