/* Copyright (C) 1997 Tom Wheeley, see file COPYING for details */

/*
 * FCAS.C  Arg checking and binary file accessing routines for .CAS files
 */

#define _CAS_FCAS_C_
 
#include "cas.h"
#include "misc.h"
#include "args.h"
#include "s9700.h"
#include "s9850.h"
#include "sraw.h"
#include "fcas.h"

static FILE *cas=NULL;

void CASReadData(char *filename, casinfo *info)
{
  darg_t dargs;
  void *hdr=NULL;

  args_join(flags.input.args,ARGS_DEF_IN,fCAS,&dargs);

  if (!(cas=tsys_fopen(filename,"rb",FALSE))) {
    Msg(EXIT_FILE,"cas/read","Unable to open `%s'",filename);
    return;
  }

  flags.input.subf=SUBF_UNKNOWN;

  if (args_is_set(dargs,"7700","9700","9800",NULL))
    flags.input.subf=SUBF_9700;
  if (args_is_set(dargs,"9750","9850","9950",NULL))
    flags.input.subf=SUBF_9850;
  if (args_is_set(dargs,"raw","uncooked",NULL))
    flags.input.subf=SUBF_RAW;

  read_datalen=0;
  if (flags.input.subf==SUBF_UNKNOWN) {
    flags.input.subf=CASReadFirst(&hdr);
  } else {
    switch(flags.input.subf) {
      case SUBF_9850: 
      	hdr=tsys_malloc(HDR_SIZE_9850);
      	CASReadBlock(hdr,HDR_SIZE_9850);
      	break;
      case SUBF_9700:
        hdr=tsys_malloc(HDR_SIZE_9700);
        CASReadBlock(hdr,HDR_SIZE_9700);
        break;
      case SUBF_RAW: ;
      default:        
    }
  }

  switch(flags.input.subf) {
    case SUBF_9700: ReadData_9700(hdr, info, &io);break;
    case SUBF_9850: ReadData_9850(hdr, info, &io);break;
    case SUBF_RAW:  ReadData_Raw (hdr, info, &io);break;
    default: Msg(EXIT_BUG,"cas/read","Unknown subfiletype `%02d'",flags.input.subf);
  }

  tsys_fclose(cas);
}


void CASWriteData(char *filename, casinfo *info)
{
  darg_t dargs;

  args_join(flags.output.args,ARGS_DEF_OUT,fCAS,&dargs);

  if (!(cas=tsys_fopen(filename,"wb",FALSE))) {
    Msg(EXIT_FILE,"cas/read","Unable to open `%s'",filename);
    return;
  }

  flags.output.subf=SUBF_UNKNOWN;
  if (flags.input.format==fCAS||flags.input.format==fCOM)
    flags.output.subf=flags.input.subf;
  if (info->type==tUnknown||info->type==tRaw)
    flags.output.subf=SUBF_RAW;

  if (args_is_set(dargs,"7700","9700","9800",NULL))
    flags.output.subf=SUBF_9700;
  if (args_is_set(dargs,"9750","9850","9950",NULL))
    flags.output.subf=SUBF_9850;
  if (args_is_set(dargs,"raw","uncooked",NULL))
    flags.output.subf=SUBF_RAW;

  switch(flags.output.subf) {
    case SUBF_9700: WriteData_9700(info, &io);break;
    case SUBF_9850: WriteData_9850(info, &io);break;
    case SUBF_RAW:  WriteData_Raw (info, &io);break;
    default: Msg(EXIT_BUG,"cas/write","Unknown subfiletype `%c'",flags.output.subf);
  }

  tsys_fclose(cas);
}


int CASReadFirst(void **hdr)
{
  unsigned char *data;
  unsigned char csum;

  data=tsys_malloc(HDR_SIZE_9850);
  (*hdr)=data;

  read_colon=getc(cas);
  if (!fread(data,1,HDR_SIZE_9700,cas)) {
    Msg(M_ERROR,"cas.read","Unable to read first header");
    free(data);
    (*hdr)=NULL;
    return 0;
  }
  read_datalen=HDR_SIZE_9700;

  if (read_colon != ':') {
    Msg(M_DEBUG,"cas/read","first: Expected `:', read `%c'",read_colon);
    read_csum=getc(cas);
    return SUBF_RAW;
  }

  if (Is_9850(data)) {                          /* 9850 */
    Msg(M_DEBUG,"cas.read","Autodetected subformat to be 9850");

    if (!fread(data + HDR_SIZE_9700,1,HDR_SIZE_9850 - HDR_SIZE_9700,cas)) {
      Msg(M_WARN,"cas.read","Unable to read extra portion of 9850 header");
    }
    read_datalen=HDR_SIZE_9850;
    read_csum=getc(cas);
    csum=CalcCSum(data,HDR_SIZE_9850);
    if (read_csum!=csum)
      Msg(M_WARN,"cas.csum","Checksum error.  expected 0x%02x, read 0x%02x",
                  csum,read_csum);

    return SUBF_9850;

  } else if (Is_9700(data)) {                    /* 9700 */
    Msg(M_DEBUG,"cas.read","Autodetected subformat to be 9700");

    read_csum=getc(cas);
    csum=CalcCSum(data,HDR_SIZE_9700);
    if (read_csum!=csum)
      Msg(M_WARN,"cas.csum","Checksum error.  expected 0x%02x, read 0x%02x",
                  csum,read_csum);

    return SUBF_9700;

  } else {                                      /* Raw */

    Msg(M_DEBUG,"cas.read","Autodetected subformat to be Raw");
    read_csum=getc(cas);
    return SUBF_RAW;
  }
}


static int CASReadBlock(void *data,size_t size)
{
  darg_t dargs;
  unsigned char csum;
  int flg_status=0;

  args_join(flags.input.args,ARGS_DEF_IN,fCAS,&dargs);

  flg_status=args_is_set(dargs,"status",NULL);

  if ((read_colon=getc(cas))!=':') {
    Msg(M_WARN,"cas/read","Expected `:', read `%c'",read_colon);
  }

  if (fread(data,1,size,cas)!=size) {
    Msg(M_ERROR,"cas/read","Unable to read 0x%04lx bytes.",size);
    return 0;
  }

  read_csum=getc(cas);
  csum=CalcCSum(data,size);
  if (read_csum!=csum)
    Msg(M_WARN,"cas/csum","Checksum error.  expected 0x%02x, read 0x%02x",
                csum,read_csum);

  if (flg_status)
    Msg(M_MSG,"cas.read","Read block of %ld (+2) bytes.",size);

  return 1;
}


static int CASWriteBlock(void *data,size_t size)
{
  darg_t dargs;
  unsigned char ccsum;
  int flg_status=0;

  args_join(flags.output.args,ARGS_DEF_OUT,fCAS,&dargs);

  flg_status=args_is_set(dargs,"status",NULL);

  putc(':',cas);
  if (fwrite(data,1,size,cas)!=size) {
    Msg(M_ERROR,"cas.write","unable to write 0x%04lx bytes",size);
    return 0;
  }
  ccsum=CalcCSum(data,size);
  putc(ccsum,cas);

  if (flg_status)
    Msg(M_MSG,"cas.write","wrote block of %ld (+2) bytes",size);

  return 1;
}


static size_t CASReadRaw(void *datablock, int trycook)
{
  darg_t dargs;
  size_t size=0;
  register char c;
  unsigned char *data;
  int flg_status=0;

  args_join(flags.input.args,ARGS_DEF_IN,fCAS,&dargs);

  flg_status=args_is_set(dargs,"status",NULL);

  data=datablock;

  read_colon='\0';      /* no chance of this being properly cooked data */
  for(;;) {
    c=getc(cas);
    if (feof(cas)) break;
    data[size++]=c;
    if (size>=CAS_MEMSIZE) {
      Msg(M_ERROR,"cas/read","File too large, cropping at %ld",size);
      break;
    }
  }

  if (flg_status)
    Msg(M_MSG,"cas.read","read file block of size %ld bytes",size);

  return size;
}


int CASWriteRaw(void *data, size_t size)
{
  darg_t dargs;

  args_join(flags.output.args,ARGS_DEF_OUT,fCAS,&dargs);

  if (args_is_set(dargs,"status",NULL)) {
    Msg(M_MSG,"cas.write","wrote file of %ld bytes",size);
  }

  return (fwrite(data,1,size,cas) != 0);
}


/* end of fcas.c */
