/* Copyright (C) 1997 Tom Wheeley, see file COPYING for details */

/*
 * FCTF.C  Routines for reading and writing Calculator Text Format files
 */

#define _CAS_FCTF_C_

#include "cas.h"
#include "misc.h"
#include "args.h"
#include "token.h"
#include "list.h"
#include "fctf.h"

void ctf_read_data(char *filename, casinfo *info)
{
  FILE *ctf=NULL;
  char *line=NULL;
  int lineno=0;
  casinfo *curblk=NULL;

  ctf = tsys_fopen(filename,"r",FALSE);
  if (!ctf) {
    Msg(M_ERROR,"ctf.read","unable to open `%s'",filename);
    return;
  }

  line = tsys_malloc(PARSE_LINESIZE);
  *line = '\0';

  /* start reading as if an @@prog header had been entered */
  curblk = info->next = tsys_malloc(sizeof (casinfo));
  curblk->type = fUnknown;
  curblk->length = 0;
  curblk->data = NULL;
  curblk->next = NULL;

  ctf_read_program(ctf, line, &lineno, curblk);

  /* check to see if our assumption was correct */
  if (curblk->length == 0) {		/* not a program */
    free(curblk);
    curblk = info;
    curblk->next = NULL;
  } else {
    curblk->type = tOldProg;
    curblk->s.prg.progtype = 0;
    curblk->s.prg.prognum = -1;
  }

  while (*line) {
    curblk->next = ctf_read_header(line, lineno);
    curblk = curblk->next;
    if (!curblk) break;
    
    switch(curblk->type) {
      case tOldProg:
      case tEditor:
        ctf_read_program(ctf, line, &lineno, curblk);
        break;
      case tRaw:
        ctf_read_raw(ctf, line, &lineno, curblk);
        break;
      case tText:
      default:
        if (curblk->type != tText)
          Msg(M_WARN,"ctf.read","unknown type, reading as plain text");
        ctf_read_text(ctf, line, &lineno, curblk);
    }
  }

  free(line);
  tsys_fclose(ctf);
}


void ctf_read_program(FILE *ctf, char *line, int *lineno, casinfo *info)
{
  int ctoken;
  unsigned char *data=NULL;
  char *lpos=NULL;
  
  info->length = 0;   /* precaution */
  data = info->data = tsys_malloc(CAS_MEMSIZE);

  in_string = insOut;

  while (ctf_read_line(ctf, line, lineno)) {
    ctf_strip_comments(line, *lineno);
    if (!line || !*line || *line == '\n') continue;

    lpos=line;

    while ((ctoken=ctf_read_token(&lpos, *lineno))) {

      /* check if approaching end of memory buffer.  We need to allow one
         more byte for the terminating 0xff */

      if (info->length >= CAS_MEMSIZE) {
        /* should be enlarging memory pool design */
        Msg(M_ERROR,"ctf.read","exhausted read buffer of size 0x%ld bytes",CAS_MEMSIZE);
        break;
      }

      switch (ctoken >> 8) {
        case 0: break;
        case 1:
          *(data++)=0x7f;   /* standard shift */
          info->length++;
          break;
        case 2:
          *(data++)=0xf7;   /* `double' shifted */
          info->length++;
          break;
        default:
          Msg(EXIT_BUG,"ctf.read","impossible token of number 0x%x",ctoken);
      }
      *(data++)=ctoken & 0xff;
      info->length++;
    }
  }

  /* display error if final char was unrecognised */
  ctf_read_token(NULL, -1);

  /* if anything was read, check for trailing newline, and add terminator */
  if (info->length) {
    if (data[-1] == PARSE_NEWLINE) {
      *(--data) = PARSE_EOF;
    } else {
      *data = PARSE_EOF;
      info->length++;
    }
  }
  
  /* shorten memory block to size required */
  info->data = tsys_realloc(info->data, info->length);
}

void ctf_read_text(FILE *ctf, char *line, int *lineno, casinfo *info)
{
  /* needs to read the data into memeory, rather than ignore it... */
  while (ctf_read_line(ctf, line, lineno))
    ;
}

void ctf_read_raw(FILE *ctf,char *line, int *lineno, casinfo *info)
{
  unsigned char *data=NULL;
  char *tmp;
  char *lpos;
  char after;
  unsigned long byte;

  info->length = 0;   /* precaution */
  data = tsys_malloc(CAS_MEMSIZE);

  while (ctf_read_line(ctf, line, lineno)) {
    if (*line == '#' || *line == ';') continue;

    /* remove any comments at start of line */
    lpos = line;
    tmp = strchr(line,' ');
    if (tmp != NULL && tmp != line) {
      if (tmp[-1] == ':') lpos=tmp;
    }
    while (isspace(*lpos))
      lpos++;

    for(;;) {
      if (!*lpos) break;

      /* reached comment part of line */
      if (isxdigit(lpos[0]) == 0) {
        Msg(M_DEBUG,"ctf.read","reached end of line: `%s'", lpos);
        break;
      }

      if (isxdigit(lpos[1]) == 0) {
        Msg(M_ERROR,"ctf.read","%d: `%c%c' is not a hexadecimal number. skipping line",
            (*lineno), lpos[0], lpos[1]);
        break;
      }

      after = lpos[2];
      if (strchr(" \t-,\n", after) == NULL) {
        Msg(M_ERROR,"ctf.read","%d, `%c' (ascii:%02x) is an illegal separator char. skipping line",
                (*lineno), after, after);
      }
      lpos[2] = '\0';

      byte = strtoul(lpos, &tmp, 16);
      if (tmp && *tmp) {
        Msg(M_WARN,"ctf.read","%d: error converting hex digits: `%s'",
                (*lineno), lpos);
      }
      if (byte > 255) {
        Msg(M_WARN,"ctf.read","%d: byte value greater than 255: `%s'",
                (*lineno), lpos);
      }
      data[info->length++] = (unsigned char) byte;

      if (!after) break;
      lpos+=3;
    }
  }

  info->data = tsys_realloc(data,info->length);
}

casinfo *ctf_read_header(char *line, int lineno)
{
  char *type=NULL;
  char *name=NULL;
  char *parm=NULL;
  char *tmp=NULL;
  casinfo *info=NULL;

  /* @@type "name" (parm) */

  if ((tmp=strrchr(line,'\n')) != NULL)  /* remove terminating newline */
    *tmp='\0';

  /* Set type to point to just after the @@ */
  type = line + strspn(line, "@ ");

  /* find beginning of parameters, put \0 to end name and set pointer to
     beginning of parm block.  then remove closing parenthesis */

  if ((parm = strchr(line,'(')) != NULL) {
    *(parm++) = '\0';
    if ((tmp = strrchr(parm,')')) != NULL) {
      *tmp = '\0';
    } else {
      Msg(M_WARN,"ctf.readhdr","%d: Unclosed parentheses",lineno);
    }
  } else {
    parm="";
  }

  /* strip spaces from around name */

  if ((name = strchr(type,' ')) != NULL) {
    while (*name == ' ')
      *(name++) = '\0';
    tmp = name + strlen(name) - 1;
    while (*tmp == ' ')
      *(tmp--) = '\0';
  } else
    name="";

  Msg(M_DEBUG,"ctf.readhdr","type=`%s', name=`%s', parm=`%s'", type, name, parm);

  info=tsys_malloc(sizeof(casinfo));
  info->type=fUnknown;
  info->length=0;
  info->data=NULL;
  info->next=NULL;

  /* allow various for program/editor.  @@prog is preferred for all */

  if (!stricmp(type, "OldProg"))
    info->type = tOldProg;
  if (!stricmp(type, "Editor") || !stricmp(type, "Edit"))
    info->type = tEditor;
  if (!stricmp(type, "Prog") || !stricmp(type, "Program")) {
    if (*name != '\"') {
      info->type = tOldProg;
    } else {
      info->type = tEditor;
    }
  }

  if (!stricmp(type, "Matrix"))
    info->type = tMatrix;
  if (!stricmp(type, "Function") || !stricmp(type, "fn"))
    info->type = tFn;
  if (!stricmp(type, "StdDev") || !stricmp(type, "sd"))
    info->type = tSD;
  if (!stricmp(type, "LinReg") || !stricmp(type, "lr"))
    info->type = tLR;

  /* special */
  if (!stricmp(type, "Desc") || !stricmp(type,"Description"))
    info->type = tDesc;
  if (!stricmp(type, "Text"))
    info->type = tText;
  if (!stricmp(type, "Raw"))
    info->type = tRaw;

  if (!stricmp(type,"Display")) {
    info->type = tText;
    Msg(M_WARN,"ctf.readhdr","%d: -l list output cannot be parsed.  reading as text", lineno);
  }

  switch (info->type) {

    case tOldProg:

      info->s.prg.prognum = -1;
      if (name && *name) {
        if (strlen(name) == 1) {
          info->s.prg.prognum = strtol(name, &tmp, 36);
          /* base 36 allows 0-9A-Z to be interpreted correctly by strtol */
          if (*tmp)
            info->s.prg.prognum = -1;
        } else {
          if (!stricmp(name, "\\r"))
            info->s.prg.prognum = 36;
          if (!strnicmp(name, "\\th",3))
            info->s.prg.prognum = 37;
                /* also allows "\theta" */
        }
      }
      if (info->s.prg.prognum < -1 || info->s.prg.prognum > 37) {
        Msg(M_ERROR,"ctf.readhdr","%d: Impossble program slot `%s'", lineno, name);
        info->s.prg.prognum=-1;
      }
      info->s.prg.progtype=0;
      if (parm && *parm) {
        for(tmp = strtok(parm," "); tmp; tmp = strtok(NULL, " ")) {
          if (!stricmp(tmp, "SD"))
            info->s.prg.progtype |= CAS_PT_SD;
          if (!stricmp(tmp, "LR"))
            info->s.prg.progtype |= CAS_PT_LR;
          if (!stricmp(tmp, "MATRIX"))
            info->s.prg.progtype |= CAS_PT_MATRIX;
          if (!stricmp(tmp, "BASE-N"))
            info->s.prg.progtype |= CAS_PT_BASEN;
          if (!stricmp(tmp, "SSTO"))
            info->s.prg.progtype |= CAS_PT_SSTO;
          if (!stricmp(tmp, "SDRAW"))
            info->s.prg.progtype |= CAS_PT_SDRAW;
        }
      }
      break;

    case tEditor:

      if (!name || !*name) {
        Msg(M_ERROR,"ctf.readhdr","%d: no program filename specified",lineno);
        free(info);
        return NULL;
      }
      if (*name == '\"')
        name++;
      if (name[strlen(name)-1] == '\"')
        name[strlen(name)-1] = '\0';
      else
        Msg(M_WARN,"ctf.readhdr","%d: Unclosed double quotes", lineno);

      strncpy(info->s.ed.filename, text_to_code(name, 12), 12);
      if (parm&&*parm)
        strncpy(info->s.ed.password, text_to_code(parm, 12), 12);
      else
        memset(info->s.ed.password, '\xff', 12);
      break;

    case tRaw:

      /* remove whitespace */
      while (*parm == ' ' || *parm == '\t')
        parm++;
      tmp = parm + strlen(parm);
      while (*tmp == ' ' || *tmp == '\t')
        *(--tmp) = '\0';

      if (!stricmp(parm,"cooked")) {
        info->s.raw.cooked = 1;
      } else {
        info->s.raw.cooked = 0;
      }
      break;

    default:
      Msg(M_ERROR,"ctf.readhdr", "%d: Unreadable file type `%s'", lineno, type);
      return NULL;
  }

  return info;
}


int ctf_read_token(char **lpos, int lineno)
{
  static char badtok[PARSE_LINESIZE];

  int ctok=0;
  char token[PARSE_TOKSIZE+1];
  char *tpos=NULL;

  /* Clearing up (called with NULL) */

  if (!lpos || !*lpos) {
    if (badtok[0] != '\0') {
      Msg(M_ERROR, "parse", "%d: unknown token or part of token `%s' (\\x%02x)",
                            lineno, badtok, *badtok);
    }
    return 0;
  }

  /* end of line */
  if (**lpos == '\0')
    return 0;

  /* remove leading spaces unless in text or comment */

#if 0
  if (in_string == insOut && (**lpos == ' ' || **lpos == '\t')) {
    while (**lpos == ' ' || **lpos == '\t')
      ++(*lpos);
  }
#endif

  /* sanity check: line ends with no \n */
  if (!**lpos) {
    Msg(M_WARN,"parse","%d: line ends on whitespace (no newline)",lineno);
    return 0;
  }
  
  /* support for `|' evaluating to `' to break up tokens */

  if (**lpos == '|') {
    (*lpos)++;
    /* tail recursion hack */
    return ctf_read_token(lpos,lineno);
  }

  /* Start with TOKSIZE size token, and reduce until valid token */

  strncpy(token, *lpos, PARSE_TOKSIZE);
  token[PARSE_TOKSIZE] = '\0';
  tpos = token + strlen(token);
  while (tpos != token) {
    ctok = token_to_code(token, 0, in_string==insOut ? 0 : TF_VERBATIM);
    if (ctok != 0) break;
    *(--tpos) = '\0'; /* reduce trial token length by 1 */
  }

  /* if in a program, ignore spaces and tabs (both mapped to ctoken 32) */
  if (in_string == insOut && ctok == PARSE_SPACE) {
    (*lpos)++;
    return ctf_read_token(lpos, lineno);
  }
  
  /* NOTE: the above code could perhaps smash the stack if there are a
   * large number of consecutive spaces in program code.
   * This should not be a problem for sensible compilers which eliminate
   * tail recursion.
   */

  /* If no token found */

  if (ctok == 0) {
    strncat(badtok, *lpos, 1); /* add token to list of bad tokens */
    (*lpos)++;
    return ctf_read_token(lpos, lineno);	/* tail recurse until good token */
  } else {
    if (*badtok) { /* if OK, then print any preceeding bad tokens */
      Msg(M_ERROR,"parse","%d: unknown token or part of token `%s' (\\x%02x)",
                        lineno,badtok,*badtok);
      *badtok = '\0';
    }
  }

  ctf_in_string_update(ctok);
  
  (*lpos)+=(tpos-token);  /* adjust current position in line */
  return ctok;
}

/* when using this function (which modifies the global variable `in_string'
 * in_string should be set to insOut to begin with.
 */

void ctf_in_string_update(int ctoken)
{
  if (ctoken == PARSE_DOUBLEQUOTE) {
    switch (in_string) {
      case insOut: 
        in_string = insString;
        break;
      case insString:
        in_string = insOut;
        break;
      case insComment:
        Msg(M_WARN,"ctf","using \" in comments may cause bugs on the calculator");
        break;
      default:
        Msg(EXIT_BUG,"ctf.in_string_update","impossible in_string value");
    }
  } else if (ctoken == PARSE_SINGLEQUOTE && in_string == insOut) {
    in_string = insComment;
  } else if ( (ctoken == PARSE_NEWLINE || ctoken == PARSE_COLON) &&
              in_string==insComment) {
    in_string = insOut;
  }
}


int ctf_read_line(FILE *ctf, char *line, int *lineno)
{
  /* add preprocessor support to this function */

  *line = '\0';
  while (*line =='\0' || *line == '\n' || *line == '#' || *line == ';') {
    /* fgets should be replaced with a function which copes better with
     * various newline conventions; ie CR, LF and CRLF
     */
    fgets(line, PARSE_LINESIZE, ctf);
    (*lineno)++;
    if (feof(ctf)) {
      *line = '\0';
      return 0;
    }
  }

  if (!strncmp(line, "@@", 2))
    return 0; /* hit header for next block */

  return 1;
}

/*
 * WARNING! This code assumes that the ' and " delimiters are single
 * characters and that their ctoken number = their ASCII value.  It also
 * assumes that no ther token includes those characters.
 */

void ctf_strip_comments(char *line, int lineno)
{
  char *tmp=NULL;
  int save_in_string = in_string;

  for (tmp = line; *tmp; tmp++) {
    ctf_in_string_update(*tmp);
    if (isspace(tmp[0]) && in_string != insString) {
      if (tmp[1] == ';' || tmp[1] == '#') {
        tmp[0]='\n';
        tmp[1]='\0';
      }
      break;
    }
  }
  in_string = save_in_string;
}


void ctf_write_data(char *filename, casinfo *info)
{
  FILE *ctf = NULL;

  ctf = tsys_fopen(filename,"w",FALSE);
  fprintf(ctf, "# Created by CaS %s\n", CAS_VER_STR);
  fprintf(ctf, "# See %s for information on tokens\n", CAS_URL);

  for(;info;info=info->next) {
    switch (info->type) {
      case tSSMono:
      case tSSCol:
        ctf_write_screenshot(ctf, info);
        break;
      case tOldProg:
      case tEditor:
        ctf_write_program(ctf, info);
        break;
      case tText:
        ctf_write_text(ctf, info);
        break;
      case tDesc:
        ctf_write_desc(ctf, info);
        break;
      case tRaw:
        ctf_write_raw(ctf, info);
        break;
      default:
        Msg(M_ERROR,"ctf.write","unsupported file type: %s",s_types[info->type]);
        break;
    }
  }

  tsys_fclose(ctf);
}

void ctf_write_program(FILE *ctf, casinfo *info)
{
  darg_t dargs;
  register unsigned int ctoken=0;
  register unsigned char *data=NULL;
  int flg_nice=0;
  
  static char *ProgramName[]={"0","1","2","3","4","5","6","7","8","9",
                              "A","B","C","D","E","F","G","H","I","J",
                              "K","L","M","N","O","P","Q","R","S","T",
                              "U","V","W","X","Y","Z","\\r","\\theta"};

  args_join(flags.output.args, ARGS_DEF_OUT,fCTF, &dargs);

  /* Write header */

  switch(info->type) {
    case tOldProg:
      fprintf(ctf, "\n@@ Program ");
      if (info->s.prg.prognum >=0 && info->s.prg.prognum <=37) {
        fprintf(ctf, "%s ", ProgramName[info->s.prg.prognum]);
      }
      fprintf(ctf, "( ");
      if (info->s.prg.progtype != 0) {
        if (info->s.prg.progtype & CAS_PT_SD)     fprintf(ctf, "SD ");
        if (info->s.prg.progtype & CAS_PT_LR)     fprintf(ctf, "LR ");
        if (info->s.prg.progtype & CAS_PT_MATRIX) fprintf(ctf, "MATRIX ");
        if (info->s.prg.progtype & CAS_PT_BASEN)  fprintf(ctf, "BASE-N ");
        if (info->s.prg.progtype & CAS_PT_SSTO)   fprintf(ctf, "SSTO ");
        if (info->s.prg.progtype & CAS_PT_SDRAW)  fprintf(ctf, "SDRAW ");
      }
      fprintf(ctf, ")\n\n");
      break;

    case tEditor:
      fprintf(ctf, "\n@@ Program \"%s\"", code_to_text(info->s.ed.filename, TOKEN_OUT_STR));
      fprintf(ctf, " (%s)\n\n", code_to_text(info->s.ed.password, TOKEN_OUT_STR));
      break;

    default:
      Msg(EXIT_BUG,"ctf.write", "ctf_write_program receievd a non-program");
  }

  /* write glossary */

  if (args_is_set(dargs, "glossary", NULL)) {
    int done_header = 0;
    int dtoken;

    /* inefficient */
    for(dtoken=0; dtoken < TOKEN_MAX; ++dtoken) {
      if (toklis->text[dtoken] == NULL)     /* check if token has text */
        continue;

      data = info->data;                    /* check if this token is */
      for(;;) {                             /* used in the program    */
        ctoken = *(data++);
        if (ctoken == 0x7f) ctoken = 256 + *(data++);
        if (ctoken == 0xf7) ctoken = 512 + *(data++);
        if (ctoken == 0xff) break;

        if (dtoken == ctoken) {             /* this token used */
          if (!done_header) {               /* write header, if not done */
            fprintf(ctf,"#  Glossary\n#\n");
            done_header=1;
          }
          fprintf(ctf,"#    %-15s %s\n", toklis->out_str[ctoken],
                                         toklis->text[ctoken]);
          break;
        }
      }
    }
    if (done_header) putc('\n',ctf);
  }

  /* write data */

  flg_nice = args_is_set(dargs, "nice", NULL);
  in_string=insOut;
  data=info->data;
  
  for (;;) {
    ctoken = *(data++);
    if (ctoken == 0x7f) {
      ctoken = 256 + *(data++);
    }
    if (ctoken == 0xf7) {
      ctoken = 512 + *(data++);
    }
    if (ctoken == 0) {
      Msg(M_DEBUG, "ctf.write", "Program terminated with 0");
      break;
    }
    if (ctoken == 255) break;

    ctf_in_string_update(ctoken);
    
    if (in_string == insOut && flg_nice != 0) {
      fprintf(ctf, "%s", code_to_token(ctoken, TOKEN_OUT_NICE));
    } else {
      fprintf(ctf, "%s", code_to_token(ctoken, TOKEN_OUT_STR));
    }
  }
  putc('\n',ctf);
}


void ctf_write_screenshot(FILE *ctf, casinfo *info)
{
  unsigned char *data;
  int x,y,col;
  /* Write header */

  switch (info->type) {
    case tSSMono:
      fprintf(ctf, "\n@@ SSMono ");
      break;
    case tSSCol:
      fprintf(ctf, "\n@@ SSCol ");
      break;
    default:
      Msg(EXIT_BUG, "ctf.write", "WriteScreenshot receievd a non-screenshot");
  }

  fprintf(ctf, "%d x %d\n", info->s.ss.width, info->s.ss.height);

  /* write data */

  fprintf(ctf, "begin\n");

  data=info->data;

  for(y=0; y<info->s.ss.height; y++) {
    for(x=0; x<info->s.ss.width; x++) {
      col = data[x+info->s.ss.width * y];
      if (info->type == tSSMono) {
        putc(col == 0 ? CTF_SS_CLEAR : CTF_SS_BLACK, ctf);
      } else {
        switch (col) {
          case 0: putc(CTF_SS_CLEAR, ctf); break;
          case 1: putc(CTF_SS_BLUE, ctf); break;
          case 2: 
          case 3: putc(CTF_SS_GREEN, ctf); break;
          case 4:
          case 5:
          case 6:
          case 7: putc(CTF_SS_RED, ctf); break;
          default: Output("x");
        }
      }
    }
    putc('\n', ctf);
  }

  fprintf(ctf, "end\n");
}


void ctf_write_raw(FILE *ctf, casinfo *info)
{
  unsigned char *data;
  int i,j,k;

  data = info->data;

  fprintf(ctf, "\n@@ Raw (");
  if (info->s.raw.cooked)
    fprintf(ctf, "cooked");
  fprintf(ctf, ")\n\n");

  for(i=0; i<info->length; i+=16) {
    k=16;if (k+i > info->length) k=info->length - i;

    /* display offset */
    fprintf(ctf, "%08x:", i);

    /* display hex data */
    for (j=0; j<k; j++)
      fprintf(ctf, "%c%02x", j==8 ? '-' : ' ', data[i+j] & 0xff);

    /* pad with spaces if neccesary */
    for (j=k; j<16; j++)
      fprintf(ctf, "   ");

    /* print ascii version */
    fprintf(ctf, "  ");
    for (j=0; j<k; j++)
      fprintf(ctf, "%s%c", j==8 ? " " : "",
              isprint(data[i+j]) ? data[i+j] : '.');
    putc('\n', ctf);
  }
}

void ctf_write_desc(FILE *ctf, casinfo *info)
{

}

void ctf_write_text(FILE *ctf, casinfo *info)
{

}

/* end of fctf.c */
