/* Copyright (C) 1997 Tom Wheeley, see file COPYING for details */

/*
 * S9850.C  Routines for translating the 9750, 9850, 9950 .CAS format
 */
 
#define _CAS_S9850_C_

#include "cas.h"
#include "misc.h"
#include "args.h"
#include "list.h"
#include "s9850.h"


void ReadData_9850(void *firsthdr, casinfo *info, iostruct *io)
{
  casheader *hdr;
  int match;
  casinfo *curblk;
  int cont=0;

  hdr=firsthdr;

  curblk=info;
  for( ; strncmp((char *)hdr->format,"END",3) ; ) {
    /* Don't accept files if their header doesn't even begin with a `:' */
    if (read_colon == ':') {
      for (match=0;match<sizeof castypes/sizeof(castypes_t);match++) {
        if (!strncmp((char *)hdr->magic,castypes[match].magic,2)) {
          cont=(*castypes[match].read)(hdr,curblk,io);
          goto read_a_block;
        }
      }
    }

    Msg(M_ERROR,"9850.read","Unable to read block of type `%c%c%c:%c%c'",
      hdr->format[0],hdr->format[1],hdr->format[2],
      hdr->magic[0],hdr->magic[1]);
    cont=1;

    read_a_block:

    if (curblk->next) curblk=curblk->next;
    if (!cont) break;
    if ((*io->readblk)(hdr, sizeof (casheader)) == 0) {
      Msg(M_WARN,"9850.read","Unable to read next header..skipping rest of file");
      free(hdr);
      return;
    }
  }

  free(hdr);
}


void WriteData_9850(casinfo *info, iostruct *io)
{
  casheader *hdr;

  for(;info;info=info->next) {
    switch(info->type) {
      case tEditor:
        if (WritePG(info,io) == 0) return;
        break;

      case tRaw:
        if (WriteRaw(info,io) == 0) return;
        break;

      case tDesc:
      case tText:
        break;
      case tEnd:
      case tUnknown:
        Msg(M_WARN,"9850.write","strange unwritable block (type: %s)",
                                            s_types[info->type]);
        break;

      default:
        Msg(M_ERROR,"9850.write","unable to write block of type: %s",
                                            s_types[info->type]);
    }
  }

  hdr=PrepareHeader("END","\xff\xff");
  memset((char *)hdr + 3,'\xff',sizeof(casheader) - 3);
  if ((*io->writeblk)(hdr,sizeof(casheader)) == 0) {
    Msg(M_WARN,"9850.write","Unable to write end marker");
  }
  free(hdr);
}

int Is_9850(unsigned char *data)
{
  int match;

  for (match=0;match<sizeof castypes/sizeof(castypes_t);match++) {
    if (!strncmp((char *)(data + 4),castypes[match].magic,2)) {
      return 1;
    }
  }
  /* no match */
  return 0;
}


/*
 * Functions to instantiate the basic data structures
 */


static casinfo *PrepareInfo(types type, size_t datalen)
{
  casinfo *info;

  info=tsys_malloc(sizeof(casinfo));

  info->type=type;
  info->length=0;       /* should only be set when data *present*  */
  info->next=NULL;
  if (datalen)
    info->data=tsys_malloc(datalen);
  else
    info->data=NULL;

  /* specific headers should be set by calling routine */

  return info;
}


static casheader *PrepareHeader(char format[3], char magic[2])
{
  int i;
  casheader *hdr;

  hdr=tsys_malloc(sizeof(casheader));

  for (i=0;i< 3;i++) hdr->format[i]=format[i];
  for (i=0;i< 2;i++) hdr->magic[i]=magic[i];
  for (i=0;i< 8;i++) {
    hdr->filename[i]='\xff';
    hdr->password[i]='\xff';
    hdr->unknown4[i]='\xff';
  }
  hdr->unknown1='\0';
  hdr->unknown2='\0';
  hdr->unknown3='\0';
  hdr->sizehi='\0';
  hdr->sizelo='\0';
  hdr->n='N';
  hdr->l='L';
  for (i=0;i<12;i++) hdr->unknown5[i]='\xff';

  return hdr;
}


/*
 * Raw data
 */


static int ReadRaw(casheader *hdr, casinfo *info, iostruct *io)
{
  casinfo *curblk;
  unsigned char *data;
  size_t datalen;

  /* copy first header into a raw block */
  curblk=info->next=PrepareInfo(tRaw,sizeof (casheader));
  curblk->s.raw.cooked=1;
  memcpy(curblk->data,hdr,sizeof (casheader));
  curblk->length=sizeof (casheader);

  /* copy all following blocks into raw blocks */

  data=tsys_malloc(CAS_MEMSIZE);
  for (;;) {
    datalen=(*io->readraw)(data, 1);

    /* screen out those annoying length=1 blocks often at end of trans */
    if (datalen<=1) {
      break;
    }

    curblk=curblk->next=PrepareInfo(tRaw,datalen);
    if (read_colon == ':') {                      /* cooked data */
      curblk->s.raw.cooked=1;
    } else {
      curblk->s.raw.cooked=0;
    }

    memcpy(curblk->data,data,datalen);
    curblk->length=datalen;
    curblk->next=NULL;
  }
  free(data);

  return 0;
}


static int WriteRaw(casinfo *info, iostruct *io)
{
  char *format;
  int headersize;

  /* If it is cooked data, then we pass it to writeblk for quicker writing
   * and to add checksum information.  If raw, or not tRaw, then send to the
   * slower writeraw, which also assumes the checksum is in the data
   */

  if (info->length > 0) {
    if (info->type == tRaw && info->s.raw.cooked) {
      if ((*io->writeblk)(info->data,info->length)==0) return 0;
    } else {
      if ((*io->writeraw)(info->data,info->length)==0) return 0;
    }
  }

  /* if the item of data is a 9850 END block, then tell the main loop
   * not to write on of it's own, then quit.
   */
  format=info->data;
  headersize=HDR_SIZE_9850;
  if (!info->s.raw.cooked) { format++; headersize+=2; }
  if (info->length==headersize &&
       !strncmp(((char *)info->data),"END",3)) {
    return 0;
  }

  /* successfully written normal block */
  return 1;
}


/*
 *  Program
 */

static int ReadPG(casheader *hdr, casinfo *info, iostruct *io)
{
  casinfo *curblk;
  long datalen;

  datalen=(hdr->sizehi << 8) + hdr->sizelo - 2;

  curblk=PrepareInfo(tEditor,datalen);

  curblk->s.ed.filename[12]='\0';
  memset(curblk->s.ed.filename,'\xff',12);
  memcpy(curblk->s.ed.filename,hdr->filename,8);

  curblk->s.ed.password[12]='\0';
  memset(curblk->s.ed.password,'\xff',12);
  memcpy(curblk->s.ed.password,hdr->password,8);

  if ((*io->readblk)(curblk->data,datalen) == 0) {
    Msg(M_WARN,"9850.read","Unable to read program data block..skipping");
    free(curblk->data);
    free(curblk);
    return 0;
  }
  curblk->length=datalen;
  info->next=curblk;

  return 1;
}

static int WritePG(casinfo *info, iostruct *io)
{
  casheader *hdr;

  hdr=PrepareHeader("TXT","PG");

  hdr->sizehi=((info->length + 2) & 0xff00) >> 8;
  hdr->sizelo=((info->length + 2) & 0x00ff);

  memcpy(hdr->filename,info->s.ed.filename,8);
  memcpy(hdr->password,info->s.ed.password,8);

  if ((*io->writeblk)(hdr,sizeof(casheader)) == 0) {
    Msg(M_WARN,"9850.write","Unable to write header..skipping block");
    free(hdr);
    return 0;
  }

  free(hdr);

  if ((*io->writeblk)(info->data,info->length) == 0) {
    Msg(M_WARN,"cas.9700.write","Unable to write data..skipping block");
    return 0;
  }

  free(hdr);
  return 1;
}


/* end of s9850.c */
