/* Copyright (C) 1997 Tom Wheeley, see file COPYING for details */

/*
 * VDRAWSS.C    Screenshot to Program convertor
 *
 *              Originally devised and written by Tom Lynn.
 *              Ported to C and integrated into CaS by Tom Wheeley
 */

#define _CAS_VDRAWSS_C_

#include "cas.h"
#include "misc.h"
#include "args.h"
#include "token.h"
#include "list.h"
#include "vdrawss.h"


void DrawSSMono(casinfo *info)
{
  static int ix[]={ 1, 0,-1, 0, 1, 1,-1,-1};
  static int iy[]={ 0, 1, 0,-1,-1, 1, 1,-1};
  int x,y;          /* current point for blanket check */
  int cx,cy;        /* current points for moving around image */
  int lx,ly;        /* line length checking points */
  int mx,my;        /* x,y co-ords of end of longest line from cx,cy */
  int dir,maxdir;   /* direction to check */
  int len,maxlen;
  unsigned char *image;
  unsigned char *notdone;
  unsigned char *program;
  unsigned char *progpos;
  
  image=info->data;
  notdone=tsys_malloc(info->s.ss.width * info->s.ss.height);
  memcpy(notdone,image,info->s.ss.width * info->s.ss.height);
  program=progpos=tsys_malloc(CAS_MEMSIZE);

  SSRange(info->s.ss.width-1,info->s.ss.height-1,&progpos);

  for (y=1;y<info->s.ss.height;++y) {
    for (x=1;x<info->s.ss.width;++x) {

      /* if point not yet plotted, then start working here */
      if (notdone[SSOFFSET(info,x,y)] == 1) {

        /* Note we must flip the y axis here due to differences between
         * CaS's internal representation and the calcs.
         */
        SSPlot(x,info->s.ss.height-y,&progpos);
        notdone[SSOFFSET(info,x,y)] = 0;
        cx=x;cy=y;

        do {
          maxlen=0;
          maxdir=0;
          mx=cx;my=cy;

          /* check length of line for each direction */
          for(dir=0;dir < 8;++dir) {
            len=0;
            lx=cx+ix[dir];
            ly=cy+iy[dir];

            for(;;) {
              /* break if outside confines of picture */
              if (lx < 1 || lx >= info->s.ss.width) break;
              if (ly < 1 || ly >= info->s.ss.height) break;

                /* we allow lines to cross areas that have already been
                 * plotted, but the line which plots the most new points
                 * is obviously the most preferable.  Hence len is increased
                 * only for points covered which have not yet been plotted
                 */

              if (image[SSOFFSET(info,lx,ly)] == 0) break;
              if (notdone[SSOFFSET(info,lx,ly)] == 1)
                ++len;

              lx+=ix[dir];
              ly+=iy[dir];
            }

            /* check if longest line so far */

            if (len > maxlen) {
              maxlen=len;
              maxdir=dir;
              mx=lx-ix[dir];
              my=ly-iy[dir];
            }
          }

          /* find a new point if no-where to move */
          if (maxlen==0) {
            Msg(M_DEBUG,"ss","nowhere to move");
            break;
          }

          /* plot line */
          SSPlot(mx,info->s.ss.height-my,&progpos);
          /* if points next to each other, then no line required */
          if (maxlen > 1) SSLine(&progpos);

          /* set points covered to done */
          for (lx=cx,ly=cy;lx!=mx || ly!=my;lx+=ix[maxdir],ly+=iy[maxdir])
            notdone[SSOFFSET(info,lx,ly)] = 0;
          notdone[SSOFFSET(info,mx,my)] = 0;

          /* set current point to cursor position */
          cx=mx;cy=my;
        } while (1);
      }
    }
  }

  SSFinish(&progpos);

  free(image);
  free(notdone);

  info->type=tOldProg;
  info->s.prg.progtype=0;
  info->s.prg.prognum=-1;
  info->data=program;
  info->length=progpos - program;
}


void SSPlot(int x, int y, unsigned char **data)
{
  static int prevx=-1;
  static int prevy=-1;

  /* We can save calc memory by using the previously plotted
   * position instead of the absolute position when possible */

  if (x==prevx && y==prevy) return;

  *((*data)++)=SS_TOK_PLOT;
  if (x==prevx && x>9) *((*data)++)='X'; else
    if (x==prevy && x>9) *((*data)++)='Y'; else 
       SSNum(x,data);
  *((*data)++)=SS_TOK_COMMA;
  if (y==prevy && y>9) *((*data)++)='Y'; else
    if (y==prevx && y>9) *((*data)++)='X'; else
      SSNum(y,data);
  *((*data)++)=SS_TOK_NL;

  prevx=x; prevy=y;
}


void SSLine(unsigned char **data)
{
  *((*data)++)=SS_TOK_LINE;
  *((*data)++)=SS_TOK_NL;
}


void SSRange(int width,int height, unsigned char **data)
{
  *((*data)++)=SS_TOK_RANGE;  *((*data)++)='1';
  *((*data)++)=SS_TOK_COMMA;  SSNum(width,data);
  *((*data)++)=SS_TOK_COMMA;  *((*data)++)='0';
  *((*data)++)=SS_TOK_COMMA;  *((*data)++)='1';
  *((*data)++)=SS_TOK_COMMA;  SSNum(height,data);
  *((*data)++)=SS_TOK_COMMA;  *((*data)++)='0';
  *((*data)++)=SS_TOK_NL;
}


void SSFinish(unsigned char **data)
{
  *((*data)++)=SS_TOK_PLOT;   *((*data)++)='0';
  *((*data)++)=SS_TOK_COMMA;  *((*data)++)='0';
  /*  *((*data)++)=SS_TOK_DISP;   Not req. at end of prog. */
  *((*data)++)=0xff;
}


void SSNum(int num, unsigned char **data)
{
  char s[20];
  char *t;

  sprintf(s,"%d",num);
  for (t=s;*t;t++)
    *((*data)++)=*t;

}

/* end of vdrawss.h */
